/*
 * Braille 'n Speak emulator
 * =========================
 *
 * ISC License
 *
 * Copyright (C) 2020 Mateusz Viste
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * ---------------------------------------------------------------------------
 *
 * Protocol implementation is based on specifications found in the Linux
 * SpeakUP driver:
 * https://github.com/torvalds/linux/blob/master/drivers/staging/speakup/speakup_bns.c
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef _WIN32
  #include <winsock2.h> /* socket() */
  #include <io.h>       /* closesocket() */
  #define CLOSESOCK(x) closesocket(x)
#else
  #include <fcntl.h>
  #include <arpa/inet.h>
  #include <sys/socket.h>
  #include <netinet/tcp.h>
  #define CLOSESOCK(x) close(x)
#endif

#include <sys/stat.h>

#include <espeak-ng/speak_lib.h>

#ifndef NOSYSLOG
#include <syslog.h>
#endif

#include "version.h"



static void logmsg(const char *s, ...) {
  va_list ap;
  va_start(ap, s);
#ifndef NOSYSLOG
  vsyslog(LOG_INFO, s, ap);
#else
  vprintf(s, ap);
  puts("");
#endif
  va_end(ap);
}


/* queue struct definition, used to store messages to be played */
struct queue_t {
  struct queue_t *next;
  char msg[1];
};


/* alloc a new queue node, add it to the tail of queue and returns new queue ptr */
static struct queue_t *add_msg_to_queue(struct queue_t *queue, const char *speechbuf, size_t speechlen) {
  struct queue_t *node;
  struct queue_t *tail;
  /* prepare new node */
  node = calloc(sizeof(*node) + speechlen + 1, 1);
  if (node == NULL) {
    logmsg("out of memory @%d", __LINE__);
    return(queue);
  }
  memcpy(node->msg, speechbuf, speechlen + 1);
  /* if queue empty, return it as-is */
  if (queue == NULL) return(node);
  /* otherwise find queue tail */
  for (tail = queue; tail->next; tail = tail->next);
  tail->next = node;
  return(queue);
}


/* free entire queue (linked list) of messages */
static void free_queue(struct queue_t *q) {
  struct queue_t *v;
  while (q) {
    v = q;
    q = q->next;
    free(v);
  }
}


/* return 0 on success */
static int parseargv(int argc, char **argv, espeak_VOICE *voicepref, unsigned short *port) {
  int i;
  *port = 7333;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-l") == 0) {
      i++;
      if (i == argc) return(-1);
      voicepref->languages = argv[i];
    } else if (strcmp(argv[i], "-v") == 0) {
      i++;
      if (i == argc) return(-1);
      voicepref->name = argv[i];
    } else if (strcmp(argv[i], "-m") == 0) {
      voicepref->gender = 1;
    } else if (strcmp(argv[i], "-f") == 0) {
      voicepref->gender = 2;
    } else if (strcmp(argv[i], "-p") == 0) {
      int p;
      i++;
      if (i == argc) return(-1);
      p = atoi(argv[i]);
      if ((p < 1) || (p > 65535)) return(-1);
      *port = (unsigned short)p;
    } else {
      return(-1);
    }
  }
  return(0);
}


/* open a listening socket on localhost:port and returns said socket */
static int listen_setup(const char *ip, unsigned short port) {
  int s;
  struct sockaddr_in serv_addr;
  /* alloc sock */
  s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (s < 0) return(-1);
  /* set up addr struct */
  memset(&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = inet_addr(ip);
  serv_addr.sin_port = htons(port);
  /* bind listening sock and switch into listen mode */
  if (bind(s, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    CLOSESOCK(s);
    logmsg("bind() failed: %s", strerror(errno));
    return(-2);
  }
  listen(s, 1);
  /* */
  return(s);
}


/* wait for a connection */
static int waitforclient(int s) {
  int clisock;
  clisock = accept(s, NULL, NULL);
  return(clisock);
}


/* set socket to non-blocking mode */
static void set_sock_nonblock(int s) {
#if _WIN32
  u_long iMode = 1;
  ioctlsocket(s, FIONBIO, &iMode); /* iMode 0=blocking 1=non-blocking */
#else
  fcntl(s, F_SETFL, fcntl(s, F_GETFL, 0) | O_NONBLOCK);
#endif
}


int main(int argc, char **argv) {
  char c;
  char buff[16] = {0};
  char speechbuf[8192] = {0};
  size_t speechlen = 0;
  int inprogress = 0;
  int sock, mastersock;
  int i;
  struct queue_t *queue = NULL;
  const espeak_VOICE **voices;
  espeak_VOICE *curvoice;
  espeak_VOICE voicepref;
  unsigned short port;

  memset(&voicepref, 0, sizeof(voicepref));

  if (parseargv(argc, argv, &voicepref, &port) != 0) {
    printf("emubns version " PVER "\n"
           "Copyright (C) 2020, Mateusz Viste\n"
           "\n"
           "emubns is a Braille 'n Speak emulator, it connects to a unix domain socket,\n"
           "awaits for BNS instructions (presumably issued by a screenreader), parses\n"
           "them and vocalize them relying on libespeak.\n"
           "\n");
    printf("emubns is published under the terms of the ISC license.\n"
           "\n"
           "usage: emubns [-p port] [-l lang] [-v voice] [-m|-f]\n"
           "\n"
           "  -p port     bind to a custom tcp port (defaults to port 7333)\n"
           "  -l lang     select language 'lang', if available (example: en-us)\n"
           "  -v voice    select voice named 'voice', if available\n"
           "  -m          prefer a male voice\n"
           "  -f          prefer a female voice\n"
           "\n");
    return(1);
  }

  /* windows require starting the WSA net engine */
#if _WIN32
  {
    WSADATA wsaData;
    WSAStartup(MAKEWORD(2,2), &wsaData);
  }
#endif

  /* prepare socket */
  mastersock = listen_setup("127.0.0.1", port);
  if (mastersock < 0) {
    logmsg("listen_setup() failed");
    return(1);
  }

  /* init libespeak */
  if (espeak_Initialize(AUDIO_OUTPUT_PLAYBACK, 0, NULL, 0) == -1) {
    CLOSESOCK(mastersock);
    fprintf(stderr, "ERROR: espeak_Initialize() failure\n");
    return(1);
  }

  /* list all available voices */
  voices = espeak_ListVoices(NULL);
  printf("available voices:\n"
         "                language | gender | name\n");
  for (i = 0; (voices != NULL) && (voices[i] != NULL); i++) {
    printf("%25s |      %c | %s\n", voices[i]->languages + 1, (voices[i]->gender == 1)?'M':'F', voices[i]->name);
  }

  /* set voice preferences */
  espeak_SetVoiceByProperties(&voicepref);

  /* display selected voice */
  curvoice = espeak_GetCurrentVoice();
  if (curvoice != NULL) {
    char *gender = "unknown gender";
    if (curvoice->gender == 1) gender = "M";
    if (curvoice->gender == 2) gender = "F";
    printf("current voice: %s / %s / %s\n", curvoice->languages + 1, curvoice->name, gender);
  } else {
    fprintf(stderr, "current voice: UNKNOWN\n");
  }

  NEXTCLIENT:

  sock = waitforclient(mastersock);
  if (sock < 0) {
    logmsg("Quitting");
    return(0);
  } else {
    logmsg("received a connection");
  }

  /* set socket to be non-blocking */
  set_sock_nonblock(sock);

  while (sock >= 0) {
    ssize_t rlen;
    /* if nothing to play, wait for activity */
    if (queue == NULL) {
      fd_set set;
      struct timeval timeout;
      FD_ZERO(&set);
      FD_SET(sock, &set);
      timeout.tv_sec = 10;
      timeout.tv_usec = 0;
      select(sock + 1, &set, NULL, NULL, &timeout);
    }
    /* read a byte from sock and possibly play a message if any is queued */
    rlen = recv(sock, &c, 1, 0);
    if (rlen == 0) {
      logmsg("connection ended");
      break;
    }
#if _WIN32
    if (rlen == SOCKET_ERROR) {
      if (WSAGetLastError() != WSAEWOULDBLOCK) {
#else
    if (rlen == -1) {
      if ((errno != EAGAIN) && (errno != EWOULDBLOCK)) {
#endif
        logmsg("connection ended with error");
        break;
      }
      /* free time - let's play something (unless playing already) */
      if ((queue != NULL) && (espeak_IsPlaying() == 0)) {
        struct queue_t *victim;
        espeak_Synth(queue->msg, strlen(queue->msg) + 1, 0, POS_CHARACTER, 0, 0, NULL, NULL);
        victim = queue;
        queue = queue->next;
        free(victim);
      }
      continue;
    }

    /* synth clear? */
    if (c == 0x18) {
      logmsg("<SYNTH CLEAR>");
      speechbuf[0]= 0;
      inprogress = 0;
      buff[0] = 0;
      free_queue(queue);
      queue = NULL;
      espeak_Cancel();
      continue;
    }

    /* start of sequence? */
    if ((inprogress == 0) && (c == 0x05)) {
      buff[0] = 0x05;
      inprogress++;
      continue;
    }

    /* seq in progress? */
    if ((inprogress > 0) && (c >= '0') && (c <= '9')) {
      /* overflow? */
      if (inprogress + 2 > (int)sizeof(buff)) {
        logmsg("SEQ OVERFLOW! '%s'", buff + 1);
        inprogress = 0;
        buff[0] = 0;
        continue;
      }
      buff[inprogress++] = c;
      buff[inprogress] = 0;
      continue;
    }

    /* seq end? */
    if ((inprogress > 0) && ((c < '0') || (c > '9'))) {
      int val = atoi(buff + 1);
      buff[inprogress++] = c;
      buff[inprogress] = 0;
      /* logmsg("GOT SEQ: '%s'", buff + 1); */
      if (c == 'P') { /* pitch: 1..20 */
        logmsg("PITCH = %d (%d)", val * 4, val);
        espeak_SetParameter(espeakPITCH, val * 4, 0);
      } else if (c == 'V') { /* volume: 1..15 */
        logmsg("VOLUME = %d (%d)", val * 13, val);
        espeak_SetParameter(espeakVOLUME, val * 13, 0);
      } else {
        logmsg("UNKNOWN SEQUENCE: %s", buff + 1);
      }
      inprogress = 0;
      buff[0] = 0;
      continue;
    }

    /* SPEAK NOW? */
    if (c == 0x0d) {
      if (speechbuf[0] != 0) {
        logmsg("TALK: '%s'", speechbuf);
        queue = add_msg_to_queue(queue, speechbuf, speechlen);
      }
      speechbuf[0] = 0;
      speechlen = 0;
      continue;
    }

    /* normal char (enqueue) */
    speechbuf[speechlen++] = c;
    speechbuf[speechlen] = 0;

  }

  free_queue(queue);
  queue = NULL;
  CLOSESOCK(sock);
  goto NEXTCLIENT;

  /* cleanup operations - never executed, but maybe useful in the future */
  /*
  espeak_Terminate();
#ifdef _WIN32
  WSACleanup();
#endif
  return(0);
  */
}
